// MainFrm.h : interface of the CMainFrame class
//
// Copyright (C) 2006 Kazuyoshi Kakihara
//
// This file is part of JanusLeaf.
//
// JanusLeaf is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_MAINFRM_H__F0355D90_C80E_4478_AD97_46A7091CA7D1__INCLUDED_)
#define AFX_MAINFRM_H__F0355D90_C80E_4478_AD97_46A7091CA7D1__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

class CMainFrame : public CFrameWindowImpl<CMainFrame>, public CUpdateUI<CMainFrame>,
		public CMessageFilter, public CIdleHandler
{
  std::string homedir;
public:
	DECLARE_FRAME_WND_CLASS(NULL, IDR_MAINFRAME)

	CMainView m_view;

	CCommandBarCtrl m_CmdBar;

	virtual BOOL PreTranslateMessage(MSG* pMsg)
  {
	  if(CFrameWindowImpl<CMainFrame>::PreTranslateMessage(pMsg))
		  return TRUE;

	  return m_view.PreTranslateMessage(pMsg);
  }

	virtual BOOL OnIdle()
  {
	  UIUpdateToolBar();
	  return FALSE;
  }

	BEGIN_UPDATE_UI_MAP(CMainFrame)
		UPDATE_ELEMENT(ID_VIEW_TOOLBAR, UPDUI_MENUPOPUP)
		UPDATE_ELEMENT(ID_VIEW_STATUS_BAR, UPDUI_MENUPOPUP)
	END_UPDATE_UI_MAP()

	BEGIN_MSG_MAP(CMainFrame)
		MESSAGE_HANDLER(WM_CREATE, OnCreate)
		COMMAND_ID_HANDLER(ID_APP_EXIT, OnFileExit)
		COMMAND_ID_HANDLER(ID_FILE_NEW, OnFileNew)
    COMMAND_ID_HANDLER(ID_FILE_OPEN, OnFileOpen)
    COMMAND_ID_HANDLER(ID_FILE_SAVE, OnFileSave)
    COMMAND_ID_HANDLER(ID_FILE_SAVE_AS, OnFileSave)
    COMMAND_ID_HANDLER(ID_FILE_EXPORTFORM, OnFileExportForm)
    COMMAND_ID_HANDLER(ID_FILE_PRINT, OnFilePrint)
    COMMAND_ID_HANDLER(ID_FILE_PRINT_PREVIEW, OnFilePrintPreview)
    COMMAND_ID_HANDLER(ID_FILE_PRINT_SETUP, OnFilePrintSetup)
    COMMAND_ID_HANDLER(ID_EDIT_UNDO, OnEditUndo)
    COMMAND_ID_HANDLER(ID_EDIT_CUT, OnEditCut)
    COMMAND_ID_HANDLER(ID_EDIT_COPY, OnEditCopy)
    COMMAND_ID_HANDLER(ID_EDIT_PASTE, OnEditPaste)
		COMMAND_ID_HANDLER(ID_VIEW_TOOLBAR, OnViewToolBar)
		COMMAND_ID_HANDLER(ID_VIEW_STATUS_BAR, OnViewStatusBar)
    COMMAND_ID_HANDLER(ID_TOOL_KEYMANAGER, OnToolKeyManager)
    COMMAND_ID_HANDLER(ID_TOOL_SHOWMYFPR, OnToolShowMyFpr)
    COMMAND_ID_HANDLER(ID_TOOL_EXPORTMYKEY, OnToolExportMyKey)
    COMMAND_ID_HANDLER(ID_HELP_CONTENTS, OnHelpContents)
		COMMAND_ID_HANDLER(ID_APP_ABOUT, OnAppAbout)
		CHAIN_MSG_MAP(CUpdateUI<CMainFrame>)
		CHAIN_MSG_MAP(CFrameWindowImpl<CMainFrame>)
	END_MSG_MAP()

// Handler prototypes (uncomment arguments if needed):
//	LRESULT MessageHandler(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
//	LRESULT CommandHandler(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
//	LRESULT NotifyHandler(int /*idCtrl*/, LPNMHDR /*pnmh*/, BOOL& /*bHandled*/)

  LRESULT OnCreate(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
  {
	  // create command bar window
	  HWND hWndCmdBar = m_CmdBar.Create(m_hWnd, rcDefault, NULL, ATL_SIMPLE_CMDBAR_PANE_STYLE);
	  // attach menu
	  m_CmdBar.AttachMenu(GetMenu());
	  // load command bar images
	  m_CmdBar.LoadImages(IDR_MAINFRAME);
	  // remove old menu
	  SetMenu(NULL);

	  HWND hWndToolBar = CreateSimpleToolBarCtrl(m_hWnd, IDR_MAINFRAME, FALSE, ATL_SIMPLE_TOOLBAR_PANE_STYLE);

	  CreateSimpleReBar(ATL_SIMPLE_REBAR_NOBORDER_STYLE);
	  AddSimpleReBarBand(hWndCmdBar);
	  AddSimpleReBarBand(hWndToolBar, NULL, TRUE);

	  CreateSimpleStatusBar();

	  //TODO: Replace with a URL of your choice
	  m_hWndClient = m_view.Create(m_hWnd, rcDefault, "about:blank", WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_HSCROLL | WS_VSCROLL, WS_EX_CLIENTEDGE);
    if (!m_hWndClient) return -1;

	  UIAddToolBar(hWndToolBar);
	  UISetCheck(ID_VIEW_TOOLBAR, 1);
	  UISetCheck(ID_VIEW_STATUS_BAR, 1);

	  // register object for message filtering and idle updates
	  CMessageLoop* pLoop = _Module.GetMessageLoop();
	  ATLASSERT(pLoop != NULL);
	  pLoop->AddMessageFilter(this);
	  pLoop->AddIdleHandler(this);

	  return 0;
  }

  LRESULT OnFileExit(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
	  PostMessage(WM_CLOSE);
	  return 0;
  }

  LRESULT OnFileNew(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.newForm();
	  return 0;
  }

  LRESULT OnFileOpen(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.loadLeaf();
	  return 0;
  }

  LRESULT OnViewToolBar(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
	  static BOOL bVisible = TRUE;	// initially visible
	  bVisible = !bVisible;
	  CReBarCtrl rebar = m_hWndToolBar;
	  int nBandIndex = rebar.IdToIndex(ATL_IDW_BAND_FIRST + 1);	// toolbar is 2nd added band
	  rebar.ShowBand(nBandIndex, bVisible);
	  UISetCheck(ID_VIEW_TOOLBAR, bVisible);
	  UpdateLayout();
	  return 0;
  }

  LRESULT OnViewStatusBar(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
	  BOOL bVisible = !::IsWindowVisible(m_hWndStatusBar);
	  ::ShowWindow(m_hWndStatusBar, bVisible ? SW_SHOWNOACTIVATE : SW_HIDE);
	  UISetCheck(ID_VIEW_STATUS_BAR, bVisible);
	  UpdateLayout();
	  return 0;
  }

  LRESULT OnHelpContents(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    CHAR szFileName[MAX_PATH + 1];
    LPSTR p;
    ::GetModuleFileName(NULL, szFileName, MAX_PATH);
    if ((p = ::PathFindFileName(szFileName)) != szFileName) {
      *p = '\0';
      if (FAILED(::StringCbCat(szFileName, MAX_PATH, "manual.chm"))) {
        return 0;
      }
      ::HtmlHelp(m_hWnd, szFileName, HH_DISPLAY_TOPIC, 0);
    }
    return 0;
  }

  LRESULT OnAppAbout(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
	  CAboutDlg dlg;
	  dlg.DoModal();
	  return 0;
  }

  LRESULT OnFileSave(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.saveLeaf();
    return 0;
  }

  LRESULT OnFilePrint(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.filePrint();
    return 0;
  }

  LRESULT OnFilePrintPreview(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.filePrintPreview();
    return 0;
  }

  LRESULT OnFilePrintSetup(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.filePrintSetup();
    return 0;
  }

  LRESULT OnEditUndo(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.editUndo();
    return 0;
  }

  LRESULT OnEditCut(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.editCut();
    return 0;
  }

  LRESULT OnEditCopy(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.editCopy();
    return 0;
  }

  LRESULT OnEditPaste(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.editPaste();
    return 0;
  }

  LRESULT OnToolKeyManager(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.toolKeyman();
    return 0;
  }

  LRESULT OnToolShowMyFpr(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.toolShowMyFpr();
    return 0;
  }

  LRESULT OnToolExportMyKey(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.toolExportMyKey();
    return 0;
  }

  LRESULT OnFileExportForm(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  {
    m_view.exportForm();
    return 0;
  }
  
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ ͑Os̒Oɒǉ̐錾}܂B

#endif // !defined(AFX_MAINFRM_H__F0355D90_C80E_4478_AD97_46A7091CA7D1__INCLUDED_)
