// MainView.h : interface of the CMainView class
//
// Copyright (C) 2006 Kazuyoshi Kakihara
//
// This file is part of JanusLeaf.
//
// JanusLeaf is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_MAINVIEW_H__2B953FDC_2D11_484F_B574_9CE5302D123E__INCLUDED_)
#define AFX_MAINVIEW_H__2B953FDC_2D11_484F_B574_9CE5302D123E__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#define KEYLEN 8

typedef struct attached_file_struct {
  UINT fileid;
  std::string filename;
  std::string tmp_filename;
  std::string passphrase;
} ATTACHED_FILE_STRUCT;

class CMainView : public CWindowImpl<CMainView, CAxWindow>, CWebExternal, ioutil
{
private:
  CComPtr<IWebBrowser2> m_pBrowser;
  std::list<ATTACHED_FILE_STRUCT> attached_files;
  std::string m_myfpr;
  std::string m_filename;
public:
  DECLARE_WND_SUPERCLASS(NULL, CAxWindow::GetWndClassName())
  CGpgParam m_gpg;

  BOOL PreTranslateMessage(MSG* pMsg)
  {
	  if((pMsg->message < WM_KEYFIRST || pMsg->message > WM_KEYLAST) &&
	     (pMsg->message < WM_MOUSEFIRST || pMsg->message > WM_MOUSELAST))
		  return FALSE;

	  // give HTML page a chance to translate this message
	  return (BOOL)SendMessage(WM_FORWARDMSG, 0, (LPARAM)pMsg);
  }

  BEGIN_MSG_MAP(CMainView)
    MESSAGE_HANDLER(WM_CREATE, OnCreate)
    MESSAGE_HANDLER(WM_DESTROY, OnDestroy)
  END_MSG_MAP()

  // Handler prototypes (uncomment arguments if needed):
  //	LRESULT MessageHandler(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/)
  //	LRESULT CommandHandler(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/)
  //	LRESULT NotifyHandler(int /*idCtrl*/, LPNMHDR /*pnmh*/, BOOL& /*bHandled*/)

  HRESULT generateEncryptKey(LPSTR buff, DWORD dwLen)
  {
    HRESULT hr = S_OK;

    _ASSERT(dwLen % 2 == 0);

    // CSP擾iȂ΍쐬j
    HCRYPTPROV hProv = NULL;
    if (!::CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, 0)) {
      if (!::CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_NEWKEYSET)) {
        hr = HRESULT_FROM_WIN32(GetLastError());
        DEBUGOUT(hr);
        hProv= NULL;
        return hr;
      }
    }

    // ꎞÍp̌쐬
    {
      LPBYTE encrypt_key_bin = new BYTE[dwLen / 2 + 1]; 
      if (!::CryptGenRandom(hProv, dwLen / 2, encrypt_key_bin)) {
        hr = HRESULT_FROM_WIN32(GetLastError());
        DEBUGOUT(hr);
        ::CryptReleaseContext(hProv, 0);
        delete [] encrypt_key_bin;
        return hr;
      }

      CHAR c;
      for (int i = 0; i < dwLen / 2; ++i) {
        c = encrypt_key_bin[i] >> 4;
        buff[i * 2] = (((c >= 0) && (c <= 9)) ? '0' + c : 'A' + c - 0xa);
        c = encrypt_key_bin[i] & 0x0f;
        buff[i * 2 + 1] = (((c >= 0) && (c <= 9)) ? '0' + c : 'A' + c - 0xa);
      }
      delete [] encrypt_key_bin;

      buff[dwLen] = '\0';

      // CSP
      if (!::CryptReleaseContext(hProv, 0)) {
        hr = HRESULT_FROM_WIN32(GetLastError());
        DEBUGOUT(hr);
        return hr;
      }
    }

    return hr;
  }

  HRESULT waitReadyStateComplete(VOID)
  {
    HRESULT hr = S_OK;
    VARIANT_BOOL bBusy = VARIANT_TRUE;
    while (TRUE) {
      m_pBrowser->get_Busy(&bBusy);
      if (bBusy == VARIANT_FALSE) break;
      MSG msg;
      if (::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
        ::TranslateMessage(&msg);
        ::DispatchMessage(&msg);
      }
    }

    READYSTATE state;
    while (TRUE) {
      m_pBrowser->get_ReadyState(&state);
      if (state == READYSTATE_COMPLETE) break;
      MSG msg;
      if (::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
        ::TranslateMessage(&msg);
        ::DispatchMessage(&msg);
      }
    }
    return hr;
  }

  LRESULT OnCreate(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM lParam, BOOL& /*bHandled*/) {
    LRESULT lr = DefWindowProc();

    m_gpg.gpgexe = "bin/gpg.exe";
    m_filename = "";

    // CXg[Ԃ̊mF
    // exet@CƓtH_INSTALLƂÕt@CWindows̃z[fBNgz[ɂ
    {
      CHAR szFileName[MAX_PATH + 1];
      LPSTR p;
      WIN32_FIND_DATA wfd;

      ::GetModuleFileName(NULL, szFileName, MAX_PATH);
      p = ::PathFindFileName(szFileName);
      *p = 0;
      if (FAILED(::StringCbCat(szFileName, MAX_PATH, "INSTALL"))) {
        return -1;
      }
      HANDLE hFind = ::FindFirstFile(szFileName, &wfd);
      if (hFind == INVALID_HANDLE_VALUE) {
        *p = 0;
        if (FAILED(::StringCbCat(szFileName, MAX_PATH, "conf"))) {
          return -1;
        }
      } else {
        ::FindClose(hFind);
        if (!::SHGetSpecialFolderPath(NULL, szFileName, CSIDL_APPDATA, FALSE)) {
          return -1;
        }
        if (FAILED(::StringCbCat(szFileName, MAX_PATH, "\\JanusLeaf\\conf"))) {
          return -1;
        }
      }
      hFind = ::FindFirstFile(szFileName, &wfd);
      if (hFind == INVALID_HANDLE_VALUE) {
        if (!::CreateDirectory(szFileName, NULL)) {
          return -1;
        }
      } else {
        ::FindClose(hFind);
      }
      m_gpg.homedir = szFileName;
      std::string seckey = m_gpg.homedir + "/secring.gpg";
      m_gpg.secring.push_back(seckey);
      std::string pubkey = m_gpg.homedir + "/pubring.gpg";
      m_gpg.pubring.push_back(pubkey);

      // ȂΌ̐
      BOOL bExist = FALSE;
      hFind = ::FindFirstFile(seckey.c_str(), &wfd);
      if (hFind == INVALID_HANDLE_VALUE) {
      } else {
        if ((wfd.nFileSizeHigh > 0) || (wfd.nFileSizeLow > 0)) {
          bExist = TRUE;
        } else {
        }
        ::FindClose(hFind);
      }
      if (!bExist) {
        CGenkeyDlg dlgGenkey;
        int nRet = dlgGenkey.DoModal(::GetActiveWindow(), reinterpret_cast<LPARAM>(&m_gpg));
        if (nRet != IDOK) return -1;
      }
    }

    // presetJǉ
    {
      CHAR szFileName[MAX_PATH + 1];
      LPSTR p;
      WIN32_FIND_DATA wfd;
      HANDLE hFind = INVALID_HANDLE_VALUE;

      ::GetModuleFileName(NULL, szFileName, MAX_PATH);
      p = ::PathFindFileName(szFileName);
      *p = 0;
      if (FAILED(::StringCbCat(szFileName, MAX_PATH, "preset"))) {
        return -1;
      }
      hFind = ::FindFirstFile(szFileName, &wfd);
      if (hFind != INVALID_HANDLE_VALUE) {
        ::FindClose(hFind);
        if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
          if (FAILED(::StringCbCat(szFileName, MAX_PATH, "/*.gpg"))) {
            return -1;
          }
          hFind = ::FindFirstFile(szFileName, &wfd);
          if (hFind != INVALID_HANDLE_VALUE) {
            m_gpg.pubring.push_back(wfd.cFileName);
            while (::FindNextFile(hFind, &wfd)) {
              m_gpg.pubring.push_back(wfd.cFileName);
            }
            ::FindClose(hFind);
          }
        }
      }
    }

    // ̌ID
    {
      std::vector<std::string> keylist;
      gpg g(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);
      g.listkey(TRUE, keylist, NULL, NULL);
      std::vector<std::string>::iterator itr = keylist.begin();
      int icol = (*itr).find(':');
      std::string strtitle = "JanusLeaf - ";
      strtitle += (*itr).substr(0, icol);
      LPCREATESTRUCT pcs = reinterpret_cast<CREATESTRUCT *>(lParam); // eEChẼnh擾
      ::SetWindowText(pcs->hwndParent, strtitle.c_str());
      m_myfpr = (*itr).substr(icol + 1);
    }

    // uEU\
    if (FAILED(this->QueryControl(&m_pBrowser))) return -1;

    // preset/index.htmlΕ\
    {
      CHAR szFileName[MAX_PATH + 1];
      LPSTR p;
      WIN32_FIND_DATA wfd;
      HANDLE hFind = INVALID_HANDLE_VALUE;

      ::GetModuleFileName(NULL, szFileName, MAX_PATH);
      p = ::PathFindFileName(szFileName);
      *p = 0;
      if (FAILED(::StringCbCat(szFileName, MAX_PATH, "preset"))) {
        return -1;
      }
      hFind = ::FindFirstFile(szFileName, &wfd);
      if (hFind != INVALID_HANDLE_VALUE) {
        ::FindClose(hFind);
        if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
          if (FAILED(::StringCbCat(szFileName, MAX_PATH, "/index.html"))) {
            return -1;
          }
          hFind = ::FindFirstFile(szFileName, &wfd);
          if (hFind != INVALID_HANDLE_VALUE) {
            ::FindClose(hFind);
            Navigate(szFileName);
          }
        }
      }
    }
    waitReadyStateComplete();
    SetExternalDispatch(static_cast<CWebExternal*>(this));
    return lr;
  }

  LRESULT OnDestroy(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& /*bHandled*/) {
    Initialize();
    return 0;
  }

  HRESULT Navigate(LPCSTR lpcFileName) {
    HRESULT hr = S_OK;

    Initialize();

    IStream* istr;
    ::CreateStreamOnHGlobal(NULL, TRUE, &istr);
    {
      HANDLE hRead;
      DWORD dwRead;
      CHAR buff[4096 + 1];
      hRead = ::CreateFile(lpcFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
      if (hRead == INVALID_HANDLE_VALUE) {
        hr = HRESULT_FROM_WIN32(::GetLastError());
        DEBUGOUT(hr);
        return hr;
      }
      while (::ReadFile(hRead, buff, 4096, &dwRead, NULL)) {
        if (dwRead == 0) break;
        if (FAILED(hr = istr->Write(buff, dwRead, NULL))) {
          ::CloseHandle(hRead);
          DEBUGOUT(hr);
          return hr;
        }    
      }
      ::CloseHandle(hRead);
    }
    if (FAILED(hr = istr->Commit(STGC_OVERWRITE))) {
      DEBUGOUT(hr);
      return hr;
    }

    LARGE_INTEGER pos = {0, 0};
    istr->Seek(pos, STREAM_SEEK_SET, NULL);
    {
      IDispatch* pDoc = NULL;
      IPersistStreamInit* pPersistStreamInit = NULL;

      // Retrieve the document object.
      if (SUCCEEDED(hr = m_pBrowser->get_Document(&pDoc))) {
        // Query for IPersistStreamInit.
        if (SUCCEEDED(hr = pDoc->QueryInterface(IID_IPersistStreamInit, (void**)&pPersistStreamInit))) {
          // Initialize the document.
          if (SUCCEEDED(hr = pPersistStreamInit->InitNew())) {
            // Load the contents of the stream.
            hr = pPersistStreamInit->Load(istr);
          }
          pPersistStreamInit->Release();
        }
      }
    }
    istr->Release();
/*
    CComVariant ve;
    CComVariant vurl(lpcFileName);
    m_pBrowser->Navigate2(&vurl, &ve, &ve, &ve, &ve);
*/
    waitReadyStateComplete();

    /*
    SetExternalDispatch(static_cast<CWebExternal*>(this));
     */
    return S_OK;
  }

  HRESULT Initialize() {
    std::list<ATTACHED_FILE_STRUCT>::iterator itr = attached_files.begin();
    while (itr != attached_files.end()) {
      ::DeleteFile((*itr).tmp_filename.c_str());
      itr = attached_files.erase(itr);
    }
    return S_OK;
  }

  HRESULT headerGetSub(std::string &headerline, LPCSTR name2, std::string &value) {
    value = "";
    std::string::iterator itr = headerline.begin();
    while (itr != headerline.end()) {
      // 󔒂̓ǂݔ΂
      while ((itr != headerline.end()) && (*itr == ' ')) ++itr;
      if (itr == headerline.end()) return S_OK; // NOT_FOUND

      // '='ʒu܂ŋ󔒂łȂǂ
      std::string::iterator itr2 = itr;
      while ((itr2 != headerline.end()) && (*itr2 != '=') && (*itr2 != ' ')) ++itr2;
      if (itr2 == headerline.end()) return HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      std::string name = std::string(itr, itr2);
      itr = itr2;
      while ((itr != headerline.end()) && (*itr == ' ')) ++itr;
      if (itr == headerline.end()) return HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      if (*itr != '=') return HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      ++itr;
      while ((itr != headerline.end()) && (*itr == ' ')) ++itr;
      itr2 = itr;
      BOOL isdq = FALSE;
      while (itr2 != headerline.end()) {
        if (isdq) {
          if (*itr2 == '"') isdq = FALSE;
        } else {
          if (*itr2 == '"') isdq = TRUE;
          else if (*itr2 == ';') break;
        }
        ++itr2;
      }
      if (isdq) return HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      if (std::string(name2) == name) {
        std::string::iterator itr3 = itr2;
        if (itr3 == itr) return S_OK; // valueȂ G[ɂׂH
        --itr3;
        while ((itr3 != itr) && (*itr3 == ' ')) --itr3;
        value = std::string(itr, itr3 + 1);
        return S_OK;
      }
      if (itr2 == headerline.end()) return S_OK; // NOT FOUND
      itr = itr2 + 1;
    }
    return S_OK; // NOT FOUND
  }

  HRESULT headerGet(std::vector<std::string> &header, LPCSTR name1, LPCSTR name2, std::string &value) {
    value = "";
    std::vector<std::string>::iterator itr;
    for (itr = header.begin(); itr != header.end(); ++itr) {
      // headername擾A̋󔒂l߁Aɓꂷ
      std::string::size_type icol;
      icol = (*itr).find(':');
      std::string::size_type ipos = icol;
      --ipos;
      while ((*itr)[ipos] == ' ') --ipos;
      std::string name = (*itr).substr(0, ipos + 1);
      int i;
      for (i = 0; i < name.size(); ++i) {
        if ((name[i] >= 'A') && (name[i] <= 'Z')) name[i] = name[i] + ('a' - 'A');
      }

      if (std::string(name1) == name) {
        std::string::size_type istart = icol + 1;
        while ((istart < (*itr).length()) && ((*itr)[istart] == ' ')) ++istart;
        BOOL isdq = FALSE;
        std::string::size_type iend = istart;
        while (iend < (*itr).length()) {
          if (isdq) {
            if ((*itr)[iend] == '"') {
              isdq = FALSE;
            }
          } else {
            if ((*itr)[iend] == '"') {
              isdq = TRUE;
            } else if ((*itr)[iend] == ';') {
              break;
            }
          }
          ++iend;
        }
        if (isdq) return HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        if (!name2) {
          while ((*itr)[iend - 1] == ' ') --iend;
          value = (*itr).substr(istart, iend - istart);
          return S_OK;
        } else {
          if (iend == (*itr).length()) return S_OK; /* HRESULT_FROM_WIN32(ERROR_NOT_FOUND); */
          istart = iend + 1;
          while ((istart < (*itr).length()) && ((*itr)[istart] == ' ')) ++istart;
          return headerGetSub((*itr).substr(istart), name2, value);
        }
      }
    }
    return S_OK; // NOT FOUND
  }

  void newForm(VOID) {
    CFileDialog dlg(TRUE, _T("html"), NULL, OFN_HIDEREADONLY | OFN_CREATEPROMPT,
      _T("HTML t@C (*.html)\0*.html\0ׂẴt@C (*.*)\0*.*\0\0"));

    if(dlg.DoModal() == IDOK){
      Navigate(dlg.m_szFileName);
      LPCSTR p = ::PathFindExtension(dlg.m_szFileName);
      if (*p == NULL) { // gq
        m_filename = dlg.m_szFileName;
        m_filename += ".gpg";
      } else {
        m_filename = dlg.m_szFileName;
        m_filename = m_filename.substr(0, p - dlg.m_szFileName) + ".gpg";
      }
    }
  }

  HRESULT loadHeader(HANDLE hRead, std::vector<std::string> &header) {
    DWORD dwRead;
    HRESULT hr = S_OK;
    std::string str, strc;
    header.clear();
    while (SUCCEEDED(hr = read(hRead, str, &dwRead))) {
      if ((dwRead == 0) || (hr == HRESULT_FROM_WIN32(ERROR_BROKEN_PIPE))) {
        // TODO: G[
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (str.length() == 0) { // wb_[I
        header.push_back(strc);
        strc = "";
        break;
      }
      if ((str.substr(0, 1) == " ") ||
          (str.substr(0, 1) == "\t")) {
        strc += str.substr(1);
      } else {
        std::string::size_type ipos = str.find(':');
        if ((ipos == std::string::npos) ||
            (ipos == 0)) {
          // TODO: G[
          hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
          DEBUGOUT(hr);
          goto finish_func;
        }
        if (strc.length() > 0) {
          header.push_back(strc);
        }
        strc = str;
      }
    }
finish_func:
    return hr;
  }
  
  HRESULT loadRawHTML(HANDLE hRead, std::string &boundary) {
    HRESULT hr = S_OK;
    IStream* istr;
    DWORD dwRead;
    std::string str, contenttype, encoding, b64buff;
    std::vector<std::string> header;
    BYTE b64decoded[4];

    if (FAILED(hr = loadHeader(hRead, header))) {
      DEBUGOUT(hr);
      goto finish_func;
    }

    if (FAILED(hr = headerGet(header, "content-type", NULL, contenttype))) {
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (contenttype != "text/html") {
      // G[
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (FAILED(hr = headerGet(header, "content-transfer-encoding", NULL, encoding))) {
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (encoding.empty()) encoding = "binary";

    CreateStreamOnHGlobal(NULL, TRUE, &istr);
    while(SUCCEEDED(hr = read(hRead, str, &dwRead))) {
      if (dwRead == 0) {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        DEBUGOUT(hr);
        goto finish_func;
      }
      if ((str.substr(0, 2) == "--") && (str.substr(2, boundary.length()) == boundary)) {
        break;
      }
      if (encoding == "BASE64") {
        b64buff += str;
        DWORD dwPos = 0, dwLen = b64buff.length(), dwBlen = 0;

        // 4oCgbase64decodeio3oCgj
        while (dwLen - dwPos >= 4) {
          dwBlen = base64decode(b64buff.c_str() + dwPos, b64decoded);
          istr->Write(b64decoded, dwBlen, NULL);
          dwPos += 4;
        }
        b64buff = b64buff.substr(dwPos);
      }
    }
    _ASSERT(b64buff.length() == 0);

    if ((dwRead == 0) || (FAILED(hr))) {
      // TODO: G[
      if (FAILED(hr)) {
        DEBUGOUT(hr);
      }
      goto finish_func;
    }

    {
      IDispatch* pDoc = NULL;
      IPersistStreamInit* pPersistStreamInit = NULL;

      istr->Commit(STGC_OVERWRITE);
      LARGE_INTEGER pos = {0, 0};
      istr->Seek(pos, STREAM_SEEK_SET, NULL);

      // Retrieve the document object.
      if (SUCCEEDED(hr = m_pBrowser->get_Document(&pDoc))) {
        // Query for IPersistStreamInit.
        if (SUCCEEDED(hr = pDoc->QueryInterface(IID_IPersistStreamInit, (void**)&pPersistStreamInit))) {
          // Initialize the document.
          if (SUCCEEDED(hr = pPersistStreamInit->InitNew())) {
            // Load the contents of the stream.
            hr = pPersistStreamInit->Load(istr);
          }
          pPersistStreamInit->Release();
        }
      }
    }
    istr->Release();

    // `҂
    waitReadyStateComplete();

finish_func:
    return hr;
  }

  HRESULT loadFilledHTML(HANDLE hRead, std::string &boundary) {
    HRESULT hr = S_OK;
    std::string str, contenttype, encoding, b64buff;
    std::vector<std::string> header;
    std::string str1;
    BYTE b64decoded[4];
    CComPtr<IDispatch> pDisp;

    if (FAILED(hr = loadHeader(hRead, header))) {
      goto finish_func;
    }

    if (FAILED(hr = headerGet(header, "content-type", NULL, contenttype))) {
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (contenttype != "application/octet-stream") {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      // TODO: G[
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (FAILED(headerGet(header, "content-transfer-encoding", NULL, encoding))) {
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (encoding.empty()) encoding = "binary";

    DWORD dwRead;
    while (SUCCEEDED(hr = read(hRead, str, &dwRead))) {
      if (dwRead == 0) {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
        DEBUGOUT(hr);
        goto finish_func;
      }
      if ((str.substr(0, 2) == "--") && (str.substr(2, boundary.length()) == boundary)) {
        break;
      }
      if (encoding == "BASE64") {
        b64buff += str;
        DWORD dwPos = 0, dwLen = b64buff.length(), dwBlen = 0;

        // 4oCgbase64decodeio3oCgj
        while (dwLen - dwPos >= 4) {
          dwBlen = base64decode(b64buff.c_str() + dwPos, b64decoded);
          dwPos += 4;
          str1 += reinterpret_cast<LPSTR>(b64decoded);
        }
        b64buff = b64buff.substr(dwPos);
      }
    }
    _ASSERT(b64buff.length() == 0);

    if ((dwRead == 0) || (FAILED(hr))) {
      // TODO: G[
      goto finish_func;
    }
/*
    {
      CComPtr<IDispatch> pDisp ;
      m_pBrowser->get_Document(&pDisp);
      CComQIPtr<IHTMLDocument2> pDoc = pDisp;

      CComQIPtr<IHTMLElement> pBody;
      pDoc->get_body(&pBody);
      CComBSTR bstr = CComBSTR(str1.c_str());
      hr = pBody->put_innerHTML(BSTR(bstr));
    }
*/
    if (SUCCEEDED(hr = m_pBrowser->get_Document(&pDisp))) {
      CComQIPtr<IHTMLDocument2> pDoc = pDisp;
      CComPtr<IDispatch> pScript;
      if (SUCCEEDED(hr = pDoc->get_Script(&pScript))) {
        CComBSTR bstrFunc("restoreForm");
        DISPID dispid = NULL;
        hr = pScript->GetIDsOfNames(IID_NULL, &bstrFunc, 1, LOCALE_SYSTEM_DEFAULT, &dispid);
        if (FAILED(hr)) { // Not Implemented
          DEBUGOUT(hr);
          return hr;
        }

        DISPPARAMS dispparams;
        ZeroMemory(&dispparams, sizeof(DISPPARAMS));
        dispparams.cArgs = 1;
        dispparams.rgvarg = new VARIANT[1];
        dispparams.cNamedArgs = 0;

        CComBSTR bstr = CComBSTR(str1.c_str());
        bstr.CopyTo(&dispparams.rgvarg[0].bstrVal);
        dispparams.rgvarg[0].vt = VT_BSTR;

        EXCEPINFO exepinfo;
        ZeroMemory(&exepinfo, sizeof(EXCEPINFO));
        CComVariant vaResult;
        UINT nArgErr = (UINT) -1;
        hr = pScript->Invoke(dispid, IID_NULL, 0, DISPATCH_METHOD, &dispparams, &vaResult, &exepinfo, &nArgErr);
        delete [] dispparams.rgvarg;
        if (FAILED(hr)) {
          DEBUGOUT(hr);
          goto finish_func;
        }
      }
    }
finish_func:
    return hr;
  }

  HRESULT loadAttachedFile(HANDLE hRead, std::string &boundary, std::string &str) {
    HRESULT hr = S_OK;
    std::string contenttype, encoding, filename, b64buff;
    std::vector<std::string> header;
    char encrypt_key[KEYLEN * 2 + 1];
    CHAR szTempFileName[MAX_PATH + 1];
    BYTE b64decoded[4];

    if (FAILED(hr = loadHeader(hRead, header))) {
      DEBUGOUT(hr);
      goto finish_func;
    }

    if (FAILED(hr = headerGet(header, "content-type", NULL, contenttype))) {
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (contenttype != "application/octet-stream") {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (FAILED(hr = headerGet(header, "content-transfer-encoding", NULL, encoding))) {
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (encoding != "BASE64") {
      hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (FAILED(hr = headerGet(header, "content-disposition", "filename", filename))) {
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (filename[0] == '"') {
      filename = filename.substr(1, filename.length() - 2);
    }

    // t@CYtf[^쐬
    {
      ATTACHED_FILE_STRUCT afs;
      CHAR szTempPath[MAX_PATH + 1];
      ::GetTempPath(MAX_PATH, szTempPath);
      ::GetTempFileName(szTempPath, "JLF", 0, szTempFileName);
      afs.fileid = attached_files.size();
      afs.filename = filename;
      afs.tmp_filename = szTempFileName;
      generateEncryptKey(encrypt_key, KEYLEN * 2);
      afs.passphrase = encrypt_key;
      attached_files.push_back(afs);
    }

    // Í
    {
      HANDLE hWrite2 = INVALID_HANDLE_VALUE, hRead2 = INVALID_HANDLE_VALUE, hFile2 = INVALID_HANDLE_VALUE;
      SECURITY_ATTRIBUTES sa;
      sa.nLength = sizeof(SECURITY_ATTRIBUTES);
      sa.bInheritHandle = TRUE;
      sa.lpSecurityDescriptor = NULL;

      gpg g2(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);
      createWritePipe(&hWrite2, &hRead2);
      hFile2 = ::CreateFile(szTempFileName, GENERIC_WRITE, 0, &sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
      if (FAILED(hr = g2.encrypt_sym_open(hRead2, hFile2, TRUE, encrypt_key))) {
        ::CloseHandle(hWrite2);
        DEBUGOUT(hr);
        goto finish_func;
      }
      hRead2 = INVALID_HANDLE_VALUE;
      hFile2 = INVALID_HANDLE_VALUE;

      DWORD dwRead;
      while (SUCCEEDED(hr = read(hRead, str, &dwRead))) {
        if (dwRead == 0) break;
        if ((str.substr(0, 2) == "--") && (str.substr(2, boundary.length()) == boundary)) {
          break;
        }
        if (encoding == "BASE64") {
          b64buff += str;
          DWORD dwPos = 0, dwLen = b64buff.length(), dwBlen = 0;

          // 4oCgbase64decodeio3oCgj
          while (dwLen - dwPos >= 4) {
            dwBlen = base64decode(b64buff.c_str() + dwPos, b64decoded);
            write(hWrite2, b64decoded, dwBlen);
            dwPos += 4;
          }
          b64buff = b64buff.substr(dwPos);
        }
      }
      _ASSERT(b64buff.length() == 0);

      if ((dwRead == 0) || (FAILED(hr))) {
        ::CloseHandle(hWrite2);
        if (FAILED(hr)) {
          DEBUGOUT(hr);
        }
        g2.close();
        goto finish_func;
      }
      ::CloseHandle(hWrite2);
      g2.close();
    }
finish_func:
    return hr;
  }

  HRESULT getPassphrase(std::string &filename, std::string &passphrase)
  {
    HRESULT hr = S_OK;
    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = TRUE;
    sa.lpSecurityDescriptor = NULL;
    HANDLE hStatusP = INVALID_HANDLE_VALUE, hStatusC = INVALID_HANDLE_VALUE;
    HANDLE hCommandP = INVALID_HANDLE_VALUE, hCommandC = INVALID_HANDLE_VALUE;
    HANDLE hRead = INVALID_HANDLE_VALUE, hNull = INVALID_HANDLE_VALUE;
    std::string str, strstatus, str1;
    DWORD dwRead;
    BOOL bSign = FALSE, bDecrypt = FALSE;
    std::vector<std::string> goodsigs;
    gpg g(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);

    hRead = CreateFile(m_filename.c_str(), GENERIC_READ, FILE_SHARE_READ, &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hRead == INVALID_HANDLE_VALUE) {
      hr = HRESULT_FROM_WIN32(::GetLastError());
      DEBUGOUT(hr);
      goto finish_func;
    }
    hNull = CreateFile("nul", GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hNull == INVALID_HANDLE_VALUE) {
      hr = HRESULT_FROM_WIN32(::GetLastError());
      DEBUGOUT(hr);
      goto finish_func;
    }

    if (FAILED(hr = createWritePipe(&hCommandP, &hCommandC))) goto finish_func;
    if (FAILED(hr = createReadPipe(&hStatusP, &hStatusC))) goto finish_func;
    hr = g.decrypt_open(hRead, hNull, TRUE, NULL, hStatusC, hCommandC);
    hRead = INVALID_HANDLE_VALUE;
    hNull = INVALID_HANDLE_VALUE;
    hStatusC = INVALID_HANDLE_VALUE;
    hCommandC = INVALID_HANDLE_VALUE;
    if (FAILED(hr)) {
      DEBUGOUT(hr);
      goto finish_func;
    }

    while (SUCCEEDED(hr = read(hStatusP, str, &dwRead))) {
      if (dwRead == 0) break;
      if (str.substr(0, 9) == "[GNUPG:] ") {
        std::string::size_type pos = str.find(' ', 9);
        if (pos == std::string::npos) {
          strstatus = str.substr(9);
        } else {
          strstatus = str.substr(9, pos - 9);
        }
      } else {
        strstatus = "";
      }
      if (strstatus == "NEED_PASSPHRASE") {
        CHAR cpassphrase[MAX_PATH + 1];
        CInputBox dlg;
        dlg.buff = cpassphrase;
        dlg.len = MAX_PATH;
        dlg.message = "閧̃bNpXt[Y͂ĂB";
        dlg.title = "pXt[Y̓";
        dlg.fPassword = TRUE;
        if (dlg.DoModal() != IDOK) {
          hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
          DEBUGOUT(hr);
          goto finish_func;
        }
        passphrase = cpassphrase;
        write(hCommandP, passphrase.c_str());
        write(hCommandP, "\n");
      } else if (strstatus == "NEED_PASSPHRASE_SYM") {
        CHAR cpassphrase[MAX_PATH + 1];
        CInputBox dlg;
        dlg.buff = cpassphrase;
        dlg.len = MAX_PATH;
        dlg.message = "Ώ̈ÍpXt[Y͂ĂB";
        dlg.title = "pXt[Y̓";
        dlg.fPassword = TRUE;
        if (dlg.DoModal() != IDOK) {
          hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
          DEBUGOUT(hr);
          goto finish_func;
        }
        passphrase = cpassphrase;
        write(hCommandP, passphrase.c_str());
        write(hCommandP, "\n");
      } else if (strstatus == "GET_LINE") {
      } else if (strstatus == "GET_HIDDEN") {
      } else if (strstatus == "GET_BOOL") {
        write(hCommandP, "y\n");
      } else if (strstatus == "VALID_SIG") {
        bSign = TRUE;
      } else if (strstatus == "GOODSIG") {
        bSign = TRUE;
        std::string::size_type pos = 0;
        pos = str.find(" "); if (pos == std::string::npos) continue;
        pos = str.find(" "); if (pos == std::string::npos) continue;
        ++pos;
        std::string strsjis;
        utf82sjis(str.substr(pos).c_str(), strsjis);
        goodsigs.push_back(strsjis);
      } else if (strstatus == "DECRYPTION_OKAY") {
        bDecrypt = TRUE;
      } else if (strstatus == "BAD_PASSPHRASE") {
        MessageBox("pXt[YɌ肪܂", "Error", MB_OK);
        hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
        DEBUGOUT(hr);
        goto finish_func;
      } else {
      }
    }

    DWORD dwExitCode;
    if (FAILED(hr = g.close(&dwExitCode))) {
      goto finish_func;
    }
    if (bDecrypt) {
      if (!bSign) {
        if (MessageBox("Lȓdq܂B𑱍s܂?", "Question", MB_YESNO) != IDYES) {
          hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
          goto finish_func;
        }
      } else {
        std::string strmsg;
        if (goodsigs.size() > 0) {
          for (std::vector<std::string>::iterator itr = goodsigs.begin(); itr != goodsigs.end(); ++itr) {
            std::string strline = *itr;
            std::string::size_type pos1, pos2;
            pos1 = strline.find(" ");
            if (pos1 == std::string::npos) goto finish_func;
            pos2 = strline.find(" ", pos1 + 1);
            if (pos2 == std::string::npos) pos2 = pos1;
            strline = strline.substr(pos2 + 1);
            strmsg += strline;
            strmsg += "\n";
          }
          strmsg += "L̓dqmF܂";
        } else {
          strmsg += "mFꂽdq͂܂";
        }
        strmsg += "\n𑱍s܂?";
        if (MessageBox(strmsg.c_str(), "Confirm", MB_YESNO) != IDYES) {
          hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
          goto finish_func;
        }
      }
    } else {
      // Error Message
      MessageBox("ł܂", "Error", MB_OK);
      hr = HRESULT_FROM_WIN32(ERROR_CANCELLED);
      goto finish_func;
    }
finish_func:
    if (hStatusP != INVALID_HANDLE_VALUE) ::CloseHandle(hStatusP);
    if (hStatusC != INVALID_HANDLE_VALUE) ::CloseHandle(hStatusC);
    if (hCommandP != INVALID_HANDLE_VALUE) ::CloseHandle(hCommandP);
    if (hCommandC != INVALID_HANDLE_VALUE) ::CloseHandle(hCommandC);
    if (hRead != INVALID_HANDLE_VALUE) ::CloseHandle(hRead);
    if (hNull != INVALID_HANDLE_VALUE) ::CloseHandle(hNull);
    return hr;
  }

  void loadLeaf(VOID) {
    USES_CONVERSION;
    DWORD dwExitCode = 0;
    Initialize();
    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = TRUE;
    sa.lpSecurityDescriptor = NULL;
    HRESULT hr = S_OK;
    HCURSOR hCursor = NULL;
    std::string passphrase;

    CFileDialog dlg(TRUE, _T("gpg"), NULL, OFN_HIDEREADONLY | OFN_CREATEPROMPT,
      _T("gpg t@C (*.gpg)\0*.gpg\0ׂẴt@C (*.*)\0*.*\0\0"));

    if(dlg.DoModal() != IDOK){
      return;
    }
    m_filename = dlg.m_szFileName;

    hCursor = ::SetCursor(::LoadCursor(NULL, IDC_WAIT));

    // ǂݍ݊mF
    hr = getPassphrase(m_filename, passphrase);
    if (FAILED(hr)) {
      ::SetCursor(hCursor);
      return;
    }

    // {ԃI[v
    HANDLE hRead = INVALID_HANDLE_VALUE, hWrite = INVALID_HANDLE_VALUE, hFile = INVALID_HANDLE_VALUE;
    std::string str, boundary;
    gpg g(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);
    createReadPipe(&hRead, &hWrite);
    hFile = ::CreateFile(m_filename.c_str(), GENERIC_READ, 0, &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE) goto finish_func;
    if (FAILED(hr = g.decrypt_open(hFile, hWrite, FALSE, passphrase.c_str()))) {
      hWrite = INVALID_HANDLE_VALUE;
      hFile = INVALID_HANDLE_VALUE;
      DEBUGOUT(hr);
      goto finish_func;
    }
    hWrite = INVALID_HANDLE_VALUE;
    hFile = INVALID_HANDLE_VALUE;

    // S̃wb_[
    {
      std::vector<std::string> header;
      std::string strct;
      if (FAILED(hr = loadHeader(hRead, header))) goto finish_func;

      // content-type`FbNAboundary擾
      if (FAILED(hr = headerGet(header, "content-type", NULL, strct))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (strct != "multipart/mixed") {
        // TODO: G[
        goto finish_func;
      }
      if (FAILED(hr = headerGet(header, "content-type", "boundary", boundary))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (*(boundary.begin()) == '"') {
        boundary = boundary.substr(1, boundary.length() - 2);
      }
    }

    // ŏboundary܂œǂݔ΂
    {
      DWORD dwRead;
      while (SUCCEEDED(hr = read(hRead, str, &dwRead))) {
        if ((dwRead == 0) || (hr == HRESULT_FROM_WIN32(ERROR_BROKEN_PIPE))) {
          hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
          DEBUGOUT(hr);
          goto finish_func;
        }
        //// I[boundaryłȂƂmF
        if (str == std::string("--") + boundary + "--") {
          // G[iŏI[͂肦Ȃj
          hr = HRESULT_FROM_WIN32(ERROR_INVALID_DATA);
          DEBUGOUT(hr);
          goto finish_func;
        } else if (str == std::string("--") + boundary) {
          break;
        }
      }
    }

    // ŏ̃ubNɂ͌eHTMLi[Ă
    if (FAILED(hr = loadRawHTML(hRead, boundary))) {
      DEBUGOUT(hr);
      goto finish_func;
    }

    // 2Ԗڂ̃ubNɂ͋LHTMLi[Ă
    if (FAILED(hr = loadFilledHTML(hRead, boundary))) {
      DEBUGOUT(hr);
      goto finish_func;
    }

    // 3Ԗڈȍ~̃ubN͓Ytt@C
    while (TRUE) {
      if (FAILED(hr = loadAttachedFile(hRead, boundary, str))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (str == std::string("--") + boundary + "--") {
        break;
      }
    }
finish_func:
    ::SetCursor(hCursor);
    if (hRead != INVALID_HANDLE_VALUE) ::CloseHandle(hRead);
    g.close();
  }

  void generateBoundary(std::string &str) {
    time_t curr;
    time(&curr);
    struct tm *newtime = localtime(&curr);
    char buff[64 + 1];

    if (FAILED(StringCbPrintf(buff, 64, "----=_%04d%02d%02d%02d%02d%02d_%04d%04d",
      newtime->tm_year,
      newtime->tm_mon,
      newtime->tm_mday,
      newtime->tm_hour,
      newtime->tm_min,
      newtime->tm_sec,
      rand(),
      rand()))) {
      str = "";
    }
    str = buff;
  }

  HRESULT saveRawHTML(HANDLE hWrite, std::string &boundary) {
    HRESULT hr = S_OK;

    write(hWrite, "--");
    write(hWrite, boundary.c_str());
    write(hWrite, "\r\n");
    write(hWrite, "Content-Type: text/html\r\n");
    write(hWrite, "Content-Transfer-Encoding: BASE64\r\n");
    write(hWrite, "\r\n");
    {
      IDispatch* pDoc = NULL;
      IPersistStreamInit* pPersistStreamInit = NULL;

      IStream* ostr;
      CreateStreamOnHGlobal(NULL, TRUE, &ostr);
      // Retrieve the document object.
      if (SUCCEEDED(hr = m_pBrowser->get_Document(&pDoc))) {
        // Query for IPersistStreamInit.
        if (SUCCEEDED(hr = pDoc->QueryInterface(IID_IPersistStreamInit, (void**)&pPersistStreamInit))) {
          // Save the document.
          hr = pPersistStreamInit->Save(ostr, TRUE);
          pPersistStreamInit->Release();
        }
      }
      LARGE_INTEGER pos = {0, 0};
      ostr->Seek(pos, STREAM_SEEK_SET, NULL);
      BYTE buff[3 + 1], buffencoded[4 + 1];
      DWORD dwRead, dwCnt = 0, dwLen = 0;
      while (ostr->Read(buff, 3, &dwRead) == S_OK) {
        if (dwRead == 0) break;
        buff[dwRead] = '\0';
        dwLen = base64encode(buff, dwRead, buffencoded);
        write(hWrite, buffencoded, dwLen);
        ++dwCnt;
        if (dwCnt == 18) {
          write(hWrite, "\r\n");
          dwCnt = 0;
        }
      }
      if (dwCnt > 0) write(hWrite, "\r\n");
      ostr->Release();
    }
    write(hWrite, "\r\n");

    return hr;
  }

  HRESULT saveFilledHTML(HANDLE hWrite, std::string &boundary) {
    HRESULT hr = S_OK;

    USES_CONVERSION;

    write(hWrite, "--");
    write(hWrite, boundary.c_str());
    write(hWrite, "\r\n");
    write(hWrite, "Content-Type: application/octet-stream\r\n");
    write(hWrite, "Content-Transfer-Encoding: BASE64\r\n");
    write(hWrite, "\r\n");

/*
    CComPtr<IDispatch> pDisp ;
    m_pBrowser->get_Document(&pDisp);
    CComQIPtr<IHTMLDocument2> pDoc = pDisp;
    CComQIPtr<IHTMLElement> pBody;
    pDoc->get_body(&pBody);
    CComBSTR bstr;
    pBody->get_innerHTML(&bstr);
    std::string str2;
    LPSTR p = OLE2T(bstr.m_str);
//    base64encode(reinterpret_cast<LPBYTE>(p), lstrlen(p), str2);
*/

    CComPtr<IDispatch> pDisp;
    if (SUCCEEDED(hr = m_pBrowser->get_Document(&pDisp))) {
      CComQIPtr<IHTMLDocument2> pDoc = pDisp;
      CComPtr<IDispatch> pScript;
      if (SUCCEEDED(hr = pDoc->get_Script(&pScript))) {
        CComBSTR bstrFunc("saveForm");
        DISPID dispid = NULL;
        hr = pScript->GetIDsOfNames(IID_NULL, &bstrFunc, 1, LOCALE_SYSTEM_DEFAULT, &dispid);
        if (FAILED(hr)) { // Not Implemented
          return hr;
        }

        DISPPARAMS dispparams;
        ZeroMemory(&dispparams, sizeof(DISPPARAMS));
        dispparams.cArgs = 0;
        dispparams.rgvarg = 0;
        dispparams.cNamedArgs = 0;

        EXCEPINFO exepinfo;
        ZeroMemory(&exepinfo, sizeof(EXCEPINFO));
        CComVariant vaResult;
        UINT nArgErr = (UINT) -1;
        hr = pScript->Invoke(dispid, IID_NULL, 0, DISPATCH_METHOD, &dispparams, &vaResult, &exepinfo, &nArgErr);
        if (vaResult.vt != VT_BSTR) goto finish_func;
        if (FAILED(hr)) goto finish_func;
        LPSTR p = OLE2A(vaResult.bstrVal);

        BYTE buffencoded[4 + 1];
        DWORD dwCnt = 0, dwLen = 0, dwPos = 0, dwLen1 = 0, dwMax = ::lstrlen(p);
        while (dwPos < dwMax - 3) {
          dwLen = base64encode(reinterpret_cast<LPCBYTE>(p) + dwPos, 3, buffencoded);
          write(hWrite, buffencoded, dwLen);
          ++dwCnt;
          if (dwCnt == 18) {
            write(hWrite, "\r\n");
            dwCnt = 0;
          }
          dwPos += 3;
        }
        dwLen = base64encode(reinterpret_cast<LPCBYTE>(p) + dwPos, dwMax - dwPos, buffencoded);
        write(hWrite, buffencoded, dwLen);
        write(hWrite, "\r\n");
      }
    }
finish_func:
    return hr;
  }

  HRESULT saveAttachedFile(HANDLE hWrite, std::string &boundary, ATTACHED_FILE_STRUCT *pafs) {
    HRESULT hr = S_OK;

    write(hWrite, "--");
    write(hWrite, boundary.c_str());
    write(hWrite, "\r\n");
    write(hWrite, "Content-Type: application/octet-stream\r\n");
    write(hWrite, "Content-Disposition: attachment;\r\n");
    write(hWrite, " filename=\"");
    write(hWrite, (*pafs).filename.c_str());
    write(hWrite, "\"\r\n");
    write(hWrite, "Content-Transfer-Encoding: BASE64\r\n");
    write(hWrite, "\r\n");

    // t@CBASE64GR[hďo
    BYTE buff[3 + 1], buffencoded[4 + 1];
    DWORD dwRead2, dwCnt = 0, dwLen = 0;

    gpg g2(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);
    HANDLE hWrite2 = INVALID_HANDLE_VALUE, hRead2 = INVALID_HANDLE_VALUE, hFile2 = INVALID_HANDLE_VALUE;
    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = TRUE;
    sa.lpSecurityDescriptor = NULL;

    if (FAILED(hr = createReadPipe(&hRead2, &hWrite2))) {
      DEBUGOUT(hr);
      goto finish_func;
    }
    hFile2 = ::CreateFile((*pafs).tmp_filename.c_str(), GENERIC_READ, 0, &sa, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile2 == INVALID_HANDLE_VALUE) {
      hr = HRESULT_FROM_WIN32(::GetLastError());
      ::CloseHandle(hRead2);
      ::CloseHandle(hWrite2);
      DEBUGOUT(hr);
      goto finish_func;
    }
    if (FAILED(hr = g2.decrypt_open(hFile2, hWrite2, FALSE, m_gpg.passphrase.c_str()))) {
      hWrite2 = INVALID_HANDLE_VALUE;
      hFile2 = INVALID_HANDLE_VALUE;
      ::CloseHandle(hRead2); hRead2 = INVALID_HANDLE_VALUE;
      DEBUGOUT(hr);
      goto finish_func;
    }
    hWrite2 = INVALID_HANDLE_VALUE;
    hFile2 = INVALID_HANDLE_VALUE;

    while (::ReadFile(hRead2, buff, 3, &dwRead2, NULL)) {
      if (dwRead2 == 0) break;
      buff[dwRead2] = '\0';
      dwLen = base64encode(buff, dwRead2, buffencoded);
      write(hWrite, buffencoded, dwLen);
      ++dwCnt;
      if (dwCnt == 18) {
        write(hWrite, "\r\n");
        dwCnt = 0;
      }
    }
    if (dwCnt > 0) write(hWrite, "\r\n");

    ::CloseHandle(hRead2);
    g2.close();

    write(hWrite, "\r\n");
finish_func:
    return hr;
  }

  void saveLeaf(void) {
    USES_CONVERSION;
    HRESULT hr = S_OK;
    HANDLE hCommandP = INVALID_HANDLE_VALUE, hCommandC = INVALID_HANDLE_VALUE;
    HANDLE hStatusP = INVALID_HANDLE_VALUE, hStatusC = INVALID_HANDLE_VALUE;
    HCURSOR hCursor = NULL;

    // o̓t@Cw
    CSelectKey dlg;
    HANDLE hWrite = INVALID_HANDLE_VALUE, hRead = INVALID_HANDLE_VALUE, hFile = INVALID_HANDLE_VALUE;
    m_gpg.filename = m_filename;
    if (dlg.DoModal(::GetActiveWindow(), reinterpret_cast<LPARAM>(&m_gpg)) == IDOK) {

      gpg g(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);
      std::vector<std::string> recipients;
      recipients = m_gpg.recipients;
      recipients.push_back(m_myfpr);

      SECURITY_ATTRIBUTES sa;
      sa.nLength = sizeof(SECURITY_ATTRIBUTES);
      sa.bInheritHandle = TRUE;
      sa.lpSecurityDescriptor = NULL;

      if (FAILED(hr = createWritePipe(&hWrite, &hRead))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      hFile = ::CreateFile(m_gpg.filename.c_str(), GENERIC_WRITE, 0, &sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
      if (hFile == INVALID_HANDLE_VALUE) {
        hr = HRESULT_FROM_WIN32(::GetLastError());
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (FAILED(hr = createWritePipe(&hCommandP, &hCommandC))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (FAILED(hr = createReadPipe(&hStatusP, &hStatusC))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (m_gpg.fPub) {
        if (FAILED(hr = g.encrypt_open(recipients, hRead, hFile, TRUE, NULL, hStatusC, hCommandC))) {
          DEBUGOUT(hr);
          goto finish_func;
        }
        hRead = INVALID_HANDLE_VALUE;
        hFile = INVALID_HANDLE_VALUE;
        hStatusC = INVALID_HANDLE_VALUE;
        hCommandC = INVALID_HANDLE_VALUE;
      } else {
        if (FAILED(g.encrypt_sym_open(hRead, hFile, TRUE, m_gpg.passphrase.c_str(), hStatusC, hCommandC))) {
          DEBUGOUT(hr);
          goto finish_func;
        }
        hRead = INVALID_HANDLE_VALUE;
        hFile = INVALID_HANDLE_VALUE;
        hStatusC = INVALID_HANDLE_VALUE;
        hCommandC = INVALID_HANDLE_VALUE;
      }

      /* gpg X^[gAbv */
      {
        std::string str, strstatus, str1;
        DWORD dwRead;
        while (SUCCEEDED(read(hStatusP, str, &dwRead))) {
          if (dwRead == 0) break;
          if (str.substr(0, 9) == "[GNUPG:] ") {
            std::string::size_type pos = str.find(' ', 9);
            if (pos == std::string::npos) {
              strstatus = str.substr(9);
            } else {
              strstatus = str.substr(9, pos - 9);
            }
          } else {
            strstatus = "";
          }
          if ((strstatus == "BEGIN_ENCRYPTION") ||
              (strstatus == "BEGIN_SIGNING")) {
            break;
          } else if (strstatus == "NEED_PASSPHRASE") {
            str1 = "閧̃bN邽߂̃pXt[Y͂ĂB";
          } else if (strstatus == "NEED_PASSPHRASE_SYM") {
            str1 = "Ώ̈Í̂߂̃pXt[Y͂ĂB";
          } else if (strstatus == "GET_LINE") {
            CInputBox dlg;
            CHAR buff[MAX_PATH + 1] = "";
            dlg.title = "Input Box";
            dlg.message = str1.c_str();
            dlg.len = MAX_PATH;
            dlg.buff = buff;
            dlg.fPassword = FALSE;
            dlg.DoModal(m_hWnd);
            if (FAILED(hr = write(hCommandP, buff))) {
              DEBUGOUT(hr);
              goto finish_func;
            }
            if (FAILED(hr = write(hCommandP, "\n"))) {
              DEBUGOUT(hr);
              goto finish_func;
            }
          } else if (strstatus == "GET_HIDDEN") {
            CInputBox dlg;
            CHAR buff[MAX_PATH + 1] = "";
            dlg.title = "Input Box";
            dlg.message = str1.c_str();
            dlg.len = MAX_PATH;
            dlg.buff = buff;
            dlg.fPassword = TRUE;
            dlg.DoModal(m_hWnd);
            if (FAILED(hr = write(hCommandP, buff))) {
              DEBUGOUT(hr);
              goto finish_func;
            }
            if (FAILED(hr = write(hCommandP, "\n"))) {
              DEBUGOUT(hr);
              goto finish_func;
            }
          } else if (strstatus == "GET_BOOL") {
            if (::MessageBox(m_hWnd, str1.c_str(), "Input Box", MB_YESNO) == IDYES) {
              if (FAILED(hr = write(hCommandP, "y\n"))) {
                DEBUGOUT(hr);
                goto finish_func;
              }
            } else {
              if (FAILED(hr = write(hCommandP, "n\n"))) {
                DEBUGOUT(hr);
                goto finish_func;
              }
            }
          } else if (strstatus == "BAD_PASSPHRASE") {
            ::MessageBox(m_hWnd, "Bad Passphrase", "Error", MB_OK);
          } else {
            str1 = str.substr(9);
          }
        }
      }

      std::string boundary;
      generateBoundary(boundary);
      hCursor = ::SetCursor(::LoadCursor(NULL, IDC_WAIT));

      // S̃wb_[o
      //      ofs << "MIME-Version: 1.0\r\n";
      if (FAILED(hr = write(hWrite, "Content-Type: multipart/mixed; boundary=\""))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (FAILED(hr = write(hWrite, boundary.c_str()))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (FAILED(hr = write(hWrite, "\"\r\n"))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (FAILED(hr = write(hWrite, "Content-Transfer-Encoding: 7bit\r\n"))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (FAILED(hr = write(hWrite, "\r\n"))) {
        DEBUGOUT(hr);
        goto finish_func;
      }

      // eHTMLo
      if (FAILED(hr = saveRawHTML(hWrite, boundary))) {
        DEBUGOUT(hr);
        goto finish_func;
      }

      // LHTMLo
      if (FAILED(hr = saveFilledHTML(hWrite, boundary))) {
        DEBUGOUT(hr);
        goto finish_func;
      }

      // Ytt@Co
      for (std::list<ATTACHED_FILE_STRUCT>::iterator itr = attached_files.begin(); itr != attached_files.end(); ++itr) {
        if (FAILED(hr = saveAttachedFile(hWrite, boundary, &(*itr)))) {
          DEBUGOUT(hr);
          goto finish_func;
        }
      }

      // I[}[J[o
      if (FAILED(hr = write(hWrite, "--"))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (FAILED(hr = write(hWrite, boundary.c_str()))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      if (FAILED(hr = write(hWrite, "--\r\n"))) {
        DEBUGOUT(hr);
        goto finish_func;
      }
      ::CloseHandle(hWrite);
      hWrite = INVALID_HANDLE_VALUE;

      ::SetCursor(hCursor);
      hCursor = NULL;

      /* gpgI */
      {
        if (hStatusP != INVALID_HANDLE_VALUE) {
          ::CloseHandle(hStatusP);
          hStatusP = INVALID_HANDLE_VALUE;
        }
        if (hCommandP != INVALID_HANDLE_VALUE) {
          ::CloseHandle(hCommandP);
          hCommandP = INVALID_HANDLE_VALUE;
        }
        DWORD dwExitCode;
        g.close(&dwExitCode);
        if (dwExitCode != 0) {
          MessageBox("Encryption Failed.", "Error", MB_OK);
        }
      }

    }
finish_func:
    if (hCursor) ::SetCursor(hCursor);
    if (hStatusP != INVALID_HANDLE_VALUE) ::CloseHandle(hStatusP);
    if (hStatusC != INVALID_HANDLE_VALUE) ::CloseHandle(hStatusC);
    if (hCommandP != INVALID_HANDLE_VALUE) ::CloseHandle(hCommandP);
    if (hCommandC != INVALID_HANDLE_VALUE) ::CloseHandle(hCommandC);
    if (hWrite != INVALID_HANDLE_VALUE) ::CloseHandle(hWrite);
    if (hRead != INVALID_HANDLE_VALUE) ::CloseHandle(hRead);
    if (hFile != INVALID_HANDLE_VALUE) ::CloseHandle(hFile);
    return;
  }

  // Ytt@C

  void attachFile(VARIANT* pvarResult) {
    USES_CONVERSION;
    CFileDialog dlg(TRUE, NULL, NULL, OFN_HIDEREADONLY | OFN_CREATEPROMPT, _T("ׂẴt@C (*.*)\0*.*\0\0"));
    if (dlg.DoModal() == IDOK) {
      ATTACHED_FILE_STRUCT afs;
      CHAR szTempPath[MAX_PATH + 1];
      ::GetTempPath(MAX_PATH, szTempPath);
      CHAR szTempFileName[MAX_PATH + 1];
      ::GetTempFileName(szTempPath, "JLF", 0, szTempFileName);

      char encrypt_key[KEYLEN * 2 + 1];
      generateEncryptKey(encrypt_key, KEYLEN * 2);

      gpg g(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);
      g.encryptfile_sym(dlg.m_szFileName, szTempFileName, FALSE, NULL, encrypt_key);
//      ::CopyFile(dlg.m_szFileName, szTempFileName, FALSE);

      afs.fileid = attached_files.size();
      afs.filename = dlg.m_szFileTitle;
      afs.tmp_filename = szTempFileName;
      afs.passphrase = encrypt_key;
      attached_files.push_back(afs);
      listFile(pvarResult);
    } else {
      BSTR bstr = A2BSTR("");
      pvarResult->vt = VT_BSTR;
      pvarResult->bstrVal = bstr;
    }
  }

  void detachFile(UINT uid, VARIANT* pvarResult) {
    UINT ui = 0;
    std::list<ATTACHED_FILE_STRUCT>::iterator itr;
    for (itr = attached_files.begin(); itr != attached_files.end();) {
      if ((*itr).fileid == uid) {
        ::DeleteFile((*itr).tmp_filename.c_str());
        itr = attached_files.erase(itr);
      } else {
        (*itr).fileid = ui;
        ++ui;
        ++itr;
      }
    }
    listFile(pvarResult);
  }

  void listFile(VARIANT* pvarResult) {
    USES_CONVERSION;

    std::string str;
    std::list<ATTACHED_FILE_STRUCT>::iterator itr;
    for (itr = attached_files.begin(); itr != attached_files.end(); ++itr) {
      char buff[MAX_PATH + 16 + 1];
      if (FAILED(StringCbPrintf(buff, MAX_PATH + 16, "%d\t%s\n", (*itr).fileid, (*itr).filename.c_str()))) {
        return;
      }
      str += buff;
    }
    pvarResult->vt = VT_BSTR;
    BSTR bstr = A2BSTR(str.c_str());
    pvarResult->vt = VT_BSTR;
    pvarResult->bstrVal = bstr;
  }

  void extractFile(UINT uid) {
    HRESULT hr = S_OK;
    for (std::list<ATTACHED_FILE_STRUCT>::iterator itr = attached_files.begin(); itr != attached_files.end(); ++itr) {
      if ((*itr).fileid == uid) {
        ATTACHED_FILE_STRUCT afs = *itr;
        CFileDialog dlg(FALSE, NULL, afs.filename.c_str(), OFN_OVERWRITEPROMPT,
                        _T("ׂẴt@C (*.*)\0*.*\0\0"));
        if (dlg.DoModal() == IDOK) {
          gpg g(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);
          if (FAILED(hr = g.decryptfile(afs.tmp_filename.c_str(), dlg.m_szFileName, FALSE, afs.passphrase.c_str()))) {
            DEBUGOUT(hr);
            return;
          }
//          ::CopyFile(afs.tmp_filename.c_str(), dlg.m_szFileName, FALSE);
          break;
        }
      }
      break;
    }
  }

  void saveForm(VARIANT var) {
    USES_CONVERSION;
    HRESULT hr = S_OK;
    CFileDialog dlg(FALSE, NULL, NULL, OFN_OVERWRITEPROMPT,
                    _T("ׂẴt@C (*.*)\0*.*\0\0"));
    if (dlg.DoModal() == IDOK) {
      HANDLE h = ::CreateFile(dlg.m_szFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
      if (h == INVALID_HANDLE_VALUE) {
        hr = HRESULT_FROM_WIN32(::GetLastError());
        DEBUGOUT(hr);
        return;
      }
      DWORD dwWritten;
      LPSTR lpstr = OLE2A(var.bstrVal);
      ::WriteFile(h, lpstr, lstrlen(lpstr), &dwWritten, NULL);
      ::CloseHandle(h);
    }
  }

  void exportForm(void) {
    HRESULT hr;
    CComPtr<IDispatch> pDisp;

    // Retrieve the document object.
    if (SUCCEEDED(hr = m_pBrowser->get_Document(&pDisp))) {
      CComQIPtr<IHTMLDocument2> pDoc = pDisp;
      CComPtr<IDispatch> pScript;
      if (SUCCEEDED(hr = pDoc->get_Script(&pScript))) {
        CComBSTR bstrFunc("exportForm");
        DISPID dispid = NULL;
        hr = pScript->GetIDsOfNames(IID_NULL, &bstrFunc, 1, LOCALE_SYSTEM_DEFAULT, &dispid);
        if (FAILED(hr)) { // Not Implemented
          return;
        }

        DISPPARAMS dispparams;
        ZeroMemory(&dispparams, sizeof(DISPPARAMS));
        dispparams.cArgs = 0;
        dispparams.rgvarg = NULL;
        dispparams.cNamedArgs = 0;

        EXCEPINFO exepinfo;
        ZeroMemory(&exepinfo, sizeof(EXCEPINFO));
        CComVariant vaResult;
        UINT nArgErr = (UINT) -1;
        hr = pScript->Invoke(dispid, IID_NULL, 0, DISPATCH_METHOD, &dispparams, &vaResult, &exepinfo, &nArgErr);
        if (FAILED(hr)) {
          DEBUGOUT(hr);
          return;
        }
        saveForm(vaResult);
      }
    }
  }

  void filePrint(void) {
    m_pBrowser->ExecWB(OLECMDID_PRINT, OLECMDEXECOPT_DODEFAULT , NULL, NULL);
  }

  void filePrintPreview(void) {
    m_pBrowser->ExecWB(OLECMDID_PRINTPREVIEW, OLECMDEXECOPT_PROMPTUSER , NULL, NULL);
  }

  void filePrintSetup(void) {
    m_pBrowser->ExecWB(OLECMDID_PAGESETUP, OLECMDEXECOPT_DODEFAULT, NULL, NULL);
  }

  void editUndo(void) {
    m_pBrowser->ExecWB(OLECMDID_UNDO, OLECMDEXECOPT_DODEFAULT, NULL, NULL);
  }

  void editCut(void) {
    m_pBrowser->ExecWB(OLECMDID_CUT, OLECMDEXECOPT_DODEFAULT, NULL, NULL);
  }

  void editCopy(void) {
    m_pBrowser->ExecWB(OLECMDID_COPY, OLECMDEXECOPT_DODEFAULT, NULL, NULL);
  }

  void editPaste(void) {
    m_pBrowser->ExecWB(OLECMDID_PASTE, OLECMDEXECOPT_DODEFAULT, NULL, NULL);
  }

  void toolKeyman(void) {
    CManageKey dlg;
    dlg.DoModal(::GetActiveWindow(), reinterpret_cast<LPARAM>(&m_gpg));
  }

  void toolShowMyFpr(void) {
    std::string strfprv;
    int ic = 0;
    for (std::string::const_iterator itr = m_myfpr.begin(); itr != m_myfpr.end(); ++itr) {
      if ((ic != 0) && (ic % 4 == 0)) strfprv.append(" ");
      if ((ic != 0) && (ic % 20 == 0)) strfprv.append("\n");
      strfprv.append(&(*itr), 1);
      ++ic;
    }
    MessageBox(strfprv.c_str(), "̌w", MB_OK);
  }

  void toolExportMyKey(void) {
    HRESULT hr = S_OK;
    CFileDialog dlg(FALSE, _T("gpg"), _T("Untitled"), OFN_OVERWRITEPROMPT,
      _T("J t@C (*.gpg)\0*.gpg\0ׂẴt@C (*.*)\0*.*\0\0"));
    if(dlg.DoModal() == IDOK) {
      gpg g(m_gpg.gpgexe.c_str(), m_gpg.homedir.c_str(), m_gpg.secring, m_gpg.pubring);
      if (FAILED(hr = g.exportkey(m_myfpr.c_str(), dlg.m_szFileName, NULL))) {
        DEBUGOUT(hr);
        return;
      }
    }
  }

};


/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ ͑Os̒Oɒǉ̐錾}܂B

#endif // !defined(AFX_MAINVIEW_H__2B953FDC_2D11_484F_B574_9CE5302D123E__INCLUDED_)
